import pandas as pd
from joblib import Parallel, delayed
from chariot.transformer.base_preprocessor import BasePreprocessor
from chariot.preprocessor import Preprocessor
from chariot.base_processor import BaseProcessor


def _apply(input, output, process, dataframe):
    input_data = dataframe[input]
    if isinstance(process, (BasePreprocessor, Preprocessor)):
        transformed = process.transform(input_data)
    else:
        transformed = process.transform(input_data.values.reshape(-1, 1))
    return (output, transformed)


class Preprocess(BaseProcessor):

    def __init__(self, spec):
        super().__init__(spec)

    @classmethod
    def _make_tasks(cls, spec):
        input_output_process = []
        for input in spec:
            if isinstance(spec[input], dict):
                for output in spec[input]:
                    process = spec[input][output]
                    input_output_process.append((input, output, process))
            else:
                process = spec[input]
                input_output_process.append((input, input, process))
        return input_output_process

    def transform(self, dataframe, n_jobs=-1, as_dataframe=False):
        tasks = self._make_tasks(self.spec)
        _targets = list(self.spec.keys())

        output_transformed = Parallel(n_jobs=n_jobs)(
            delayed(_apply)(t[0], t[1], t[2], dataframe) for t in tasks)

        applied = {}
        for o_t in output_transformed:
            output, transformed = o_t
            applied[output] = transformed

        for c in dataframe.columns:
            if c not in _targets:
                applied[c] = dataframe[c]

        if not as_dataframe:
            return applied
        else:
            applied = pd.DataFrame.from_dict(applied)
            return applied



import os
import shutil
import json
from pathlib import Path
import tarfile
from sklearn.externals import joblib


class BaseProcessor():

    def __init__(self, spec):
        self.spec = spec

    def save(self, path):
        basename = os.path.basename(path).split(".")[0]
        root = Path(os.path.dirname(path)).joinpath(basename)
        root.mkdir()

        keymap = {}
        for k in self.spec:
            if isinstance(self.spec[k], dict):
                keymap[k] = {}
                for j in self.spec[k]:
                    keymap[k][j] = "_".join([k, j]) + ".pkl"
                    joblib.dump(self.spec[k][j], root.joinpath(keymap[k][j]))
            else:
                keymap[k] = k + ".pkl"
                joblib.dump(self.spec[k], root.joinpath(keymap[k]))

        with root.joinpath("spec.json").open(mode="w", encoding="utf-8") as f:
            json.dump(keymap, f)

        with tarfile.open(path, "w:gz") as tar:
            tar.add(root, arcname=basename)

        shutil.rmtree(root)

    @classmethod
    def load(cls, path):
        basename = os.path.basename(path).split(".")[0]
        with tarfile.open(path, "r:gz") as tar:
            spec_f = tar.extractfile(basename + "/spec.json")
            spec = json.load(spec_f)
            for k in spec:
                if isinstance(spec[k], dict):
                    for j in spec[k]:
                        f_name = "_".join([k, j]) + ".pkl"
                        f_name = basename + "/" + f_name
                        p = joblib.load(tar.extractfile(f_name))
                        spec[k][j] = p
                else:
                    f_name = k + ".pkl"
                    f_name = basename + "/" + f_name
                    spec[k] = joblib.load(tar.extractfile(f_name))

        instance = cls(spec)
        return instance

    def inverse_transform(self, X):
        self._validate_transformers()
        Xt = X
        for name, t in list(self._transformers)[::-1]:
            if hasattr(t, "inverse_transform"):
                Xt = t.inverse_transform(Xt)
        return Xt



import pandas as pd
import numpy as np


class TargetGenerator():

    def __init__(self):
        pass

    def generate(self, data, index, length):
        raise Exception("You should implements generate in subclass.")


class ShiftGenerator(TargetGenerator):

    def __init__(self, shift=1):
        super().__init__()
        self.shift = shift

    def generate(self, data, index, length):
        _from = data[index:index+length]
        _to = data[index+self.shift:index+self.shift+length]
        return _from, _to


class ShuffleGenerator(TargetGenerator):

    def __init__(self):
        super().__init__()

    def generate(self, data, index, length):
        _from = data[index:index+length]
        _to = np.apply_along_axis(np.random.permutation, 1, _from)
        return _from, _to



import pandas as pd
import numpy as np
from chariot.transformer.formatter.base import BaseFormatter
from chariot.preprocessor import Preprocessor


class Padding(BaseFormatter):

    def __init__(self, padding=0, length=-1,
                 begin_of_sequence=-1, end_of_sequence=-1):
        super().__init__()
        self.padding = padding
        self.length = length
        self.begin_of_sequence = begin_of_sequence
        self.end_of_sequence = end_of_sequence

    @classmethod
    def from_(cls, vocabulary_or_preprocessor, length=-1,
              begin_of_sequenceuence=False, end_of_sequenceuence=False):

        vocabulary = vocabulary_or_preprocessor
        if isinstance(vocabulary_or_preprocessor, Preprocessor):
            vocabulary = vocabulary_or_preprocessor.vocabulary

        padding = Padding(padding=vocabulary.pad, length=length)
        if begin_of_sequenceuence:
            padding.begin_of_sequence = vocabulary.bos
        if end_of_sequenceuence:
            padding.end_of_sequence = vocabulary.eos

        return padding

    def transform(self, column):
        def adjust(sequence, length):
            if self.begin_of_sequence > 0:
                sequence = [self.begin_of_sequence] + sequence
            if self.end_of_sequence > 0:
                sequence = sequence + [self.end_of_sequence]

            if self.length > 0:
                if len(sequence) < self.length:
                    pad_size = self.length - len(sequence)
                    sequence = sequence + [self.padding] * pad_size
                elif len(sequence) > self.length:
                    sequence = sequence[:length]

            return np.array(sequence)

        length = self.length
        if length < 0:
            length = max([len(seq) for seq in column])

        if isinstance(column, pd.Series):
            return column.apply(lambda x: adjust(x, length))
        else:
            return np.array([adjust(x, length) for x in column])

    def inverse_transform(self, column):
        def inverse(sequence):
            excludes = [self.padding,
                        self.begin_of_sequence, self.end_of_sequence]
            inversed = [t for t in sequence if t not in excludes]
            return inversed

        if isinstance(column, pd.Series):
            return column.apply(lambda x: inverse(x))
        else:
            return np.array([inverse(x) for x in column])



import re
import copy
import numpy as np
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.base import BaseEstimator, TransformerMixin
from chariot.transformer.vocabulary import Vocabulary
from chariot.transformer.tokenizer import Tokenizer


class Preprocessor(_BaseComposition, BaseEstimator, TransformerMixin):

    def __init__(self, tokenizer=None,
                 text_transformers=(), token_transformers=(),
                 vocabulary=None):
        self.tokenizer = tokenizer
        if isinstance(self.tokenizer, str):
            self.tokenizer = Tokenizer(self.tokenizer)
        self.text_transformers = text_transformers
        self.token_transformers = token_transformers
        self.vocabulary = vocabulary

    def _to_snake(self, name):
        _name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        _name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", _name).lower()
        return _name

    @property
    def _transformers(self):
        transformers = list(self.text_transformers)
        if self.tokenizer:
            transformers += [self.tokenizer]
        transformers += self.token_transformers
        if self.vocabulary:
            transformers += [self.vocabulary]

        return (
            (self._to_snake(t.__class__.__name__) + "_", t) for t in
            transformers
        )

    """
    @_transformers.setter
    def _transformers(self, value):
        print("XXXXXXXXXXXXXXXXX")
        for name, t in value:
            if isinstance(t, (TextFilter, TextNormalizer)):
                self.text_transformers.append(t)
            elif isinstance(t, (TokenFilter, TokenNormalizer)):
                self.token_transformers.append(t)
            elif isinstance(t, Tokenizer):
                self.tokenizer = t
            elif isinstance(t, Vocabulary):
                self.vocabulary = t

    def get_params(self, deep=True):
        return self._get_params("_transformers", deep=deep)

    def set_params(self, **kwargs):
        self._set_params("_transformers", **kwargs)
        return self
    """

    def _validate_transformers(self):
        names, transformers = zip(*self._transformers)

        # validate names
        self._validate_names(names)

        # validate estimators
        for t in transformers:
            if t is None:
                continue
            if (not (hasattr(t, "fit") or hasattr(t, "fit_transform")) or not
                    hasattr(t, "transform")):
                raise TypeError("All transformer should implement fit and "
                                "transform. '%s' (type %s) doesn't" %
                                (t, type(t)))

    def transform(self, X):
        self._validate_transformers()
        Xt = self.check_array(X, True)
        for name, t in self._transformers:
            original_copy_setting = t.copy
            t.copy = False  # Don't copy in each transformer
            Xt = t.transform(Xt)
            t.copy = original_copy_setting
        return Xt

    def inverse_transform(self, X):
        self._validate_transformers()
        Xt = X
        for name, t in list(self._transformers)[::-1]:
            if hasattr(t, "inverse_transform"):
                Xt = t.inverse_transform(Xt)
        return Xt

    def fit(self, X, y=None):
        self._validate_transformers()
        Xt = X
        for name, t in self._transformers:
            if isinstance(t, Vocabulary):
                t.fit(Xt)
            else:
                Xt = t.fit_transform(Xt)
        return self

    def fit_transform(self, X, y=None):
        self._validate_transformers()
        Xt = X
        for name, t in self._transformers:
            Xt = t.fit_transform(Xt)
        return Xt

    def check_array(self, X, copy_obj=True):
        if not copy_obj:
            return X
        else:
            if isinstance(X, (np.ndarray, np.generic)):
                return np.array(X)
            else:
                return copy.deepcopy(X)



import pandas as pd
import numpy as np


def apply_map(data, func, inplace=False):

    if isinstance(data, pd.DataFrame):
        # apply and apply_map call function twice.
        result = data if inplace else {}
        for c in data.columns:
            s = list(map(func, data[c].values))
            if inplace:
                result.loc[:, c] = s
            else:
                result[c] = s
        if not inplace:
            result = pd.DataFrame.from_dict(result)
        return result
    elif isinstance(data, pd.Series):
        if inplace:
            for i, e in enumerate(map(func, data.values)):
                data.iat[i] = e
            return data
        else:
            result = pd.Series(list(map(func, data.values)))
            return result
    elif isinstance(data, dict):
        result = data if inplace else {}
        for k in list(data.keys()):
            result[k] = [func(row) for row in data[k]]
        return result
    elif isinstance(data, (list, tuple)):
        result = data if inplace else ([0] * len(data))

        # 2d array
        if len(data) > 0 and isinstance(data[0], (list, tuple)):
            data_dim = len(np.array(data[0]).shape)
            for i, row in enumerate(data):
                if data_dim == 1:
                    result[i] = func(row)
                else:
                    # row has multiple column
                    _row = []
                    for column in row:
                        _row.append(func(column))
                    result[i] = _row
        else:
            for i, e in enumerate(data):
                result[i] = func(e)
        return result
    else:
        return func(data)

    return data



import itertools
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from chariot.base_processor import BaseProcessor
from chariot.transformer.formatter.target_generator import TargetGenerator


def _apply(target, process, data, inverse=False):
    input_data = data[target]
    if isinstance(input_data, pd.Series):
        input_data = input_data.values

    if not inverse:
        transformed = process.transform(input_data)
    else:
        if hasattr(process, "inverse_transform"):
            transformed = process.inverse_transform(input_data)
    return (target, transformed)


class Feeder(BaseProcessor):

    def __init__(self, spec):
        super().__init__(spec)
        self._resource = None

    def set_resource(self, data):
        self._resource = data
        return self

    def _get_keys(self, data, ignores):
        if isinstance(data, pd.DataFrame):
            keys = [c for c in data.columns if c not in ignores]
        else:
            keys = [k for k in data if k not in ignores]
        return keys

    def transform(self, data=None, ignores=(), n_jobs=-1, inverse=False):
        _data = data if data else self._resource
        tasks = []
        for k in self.spec:
            if k not in ignores:
                tasks.append((k, self.spec[k]))

        target_transformed = Parallel(n_jobs=n_jobs)(
            delayed(_apply)(i, p, _data, inverse) for i, p in tasks)
        target_transformed = dict(target_transformed)

        applied = {}
        for key in self._get_keys(_data, ignores):
            if key in target_transformed:
                applied[key] = target_transformed[key]
            else:
                applied[key] = _data[key]

        return applied

    def make_generator(self, data, batch_size, epoch=-1, n_jobs=-1,
                       ignores=()):
        _data = data if data else self._resource
        data_length = len(_data)
        steps_per_epoch = data_length // batch_size
        keys = self._get_keys(_data, ignores)

        def generator():
            indices = np.arange(data_length)
            count = 0
            _epoch = 0
            while True:
                if count > 0 and count % steps_per_epoch == 0:
                    _epoch += 1
                    count = 0
                    if epoch > 0 and _epoch >= epoch:
                        break

                selected = np.random.choice(indices, size=batch_size,
                                            replace=False)

                if isinstance(_data, pd.DataFrame):
                    batch = _data[keys].iloc[selected, :]
                else:
                    batch = {}
                    for key in keys:
                        if isinstance(_data[key], pd.Series):
                            batch[key] = _data[key].iloc[selected]
                        else:
                            batch[key] = _data[key][selected]

                count += 1
                yield self.transform(batch, n_jobs=n_jobs)

        return steps_per_epoch, generator

    def iterate(self, data=None, batch_size=32, epoch=-1, n_jobs=-1,
                ignores=()):
        _, generator = self.make_generator(data, batch_size, epoch, n_jobs,
                                           ignores)
        for b in generator():
            yield b

    def inverse_transform(self, batch, n_jobs=-1, ignores=()):
        return self.transform(batch, n_jobs=n_jobs, ignores=ignores, inverse=True)


class LanguageModelFeeder(BaseProcessor):

    def __init__(self, spec):
        super().__init__(spec)
        self._resource = None
        self.key = ""

        if len(spec) != 1:
            raise Exception("Language model should have only 1 column\
                            that contains sentences.")
        else:
            self.key = list(spec.keys())[0]
            if not isinstance(self.spec[self.key], TargetGenerator):
                raise Exception("Language model should take one generator.")

    def set_resource(self, data, batch_size):
        sentences = data[self.key]
        if isinstance(sentences, pd.Series):
            sentences = sentences.values
        elif isinstance(sentences, (list, tuple)):
            np.array(sentences)

        self._resource = np.array(list(itertools.chain.from_iterable(sentences)))
        limit = (len(self._resource) // batch_size) * batch_size
        self._resource = self._resource[:limit].reshape((batch_size, -1)).T

    def make_generator(self, data, batch_size, sequence_length, epoch=-1,
                       sequencial=True):

        self.set_resource(data, batch_size)
        steps_per_epoch = (len(self._resource) - 1) // sequence_length
        _generator = self.spec[self.key]

        def generator():
            count = 0
            _epoch = 0
            while True:
                if count > 0 and count % steps_per_epoch == 0:
                    _epoch += 1
                    count = 0
                    if epoch > 0 and _epoch >= epoch:
                        break
                
                index = count * sequence_length
                data, target = _generator.generate(self._resource, index, sequence_length)
                count += 1
                if sequencial:
                    yield data, target
                else:
                    # Predict one word from sequence
                    yield data.T, target[-1].T

        return steps_per_epoch, generator

    def iterate(self, data, batch_size, sequence_length, epoch=-1, sequencial=True):
        _, generator = self.make_generator(data, batch_size, sequence_length, epoch, sequencial)
        for d, t in generator():
            yield d, t
